#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define ALL MPI_COMM_WORLD
#define N 8               /* Fixed value */

void main(int argc, char **argv) {

  /* Integer, Double type variable declaration */
  int size, id, i, j, k, n;
  int leftnb, rightnb;
  double *x, *y, *correlation, buffer;

  /* MPI_Status declaration */
  MPI_Status status;

  /* Initialize MPI functions */
  MPI_Init(&argc, &argv);

  /* In order to find the number of processes */
  MPI_Comm_size(ALL, &size);

  /* In order for a process to find out its rank */
  MPI_Comm_rank(ALL, &id);

  /* Number to allocate for each process  */
  n = N / size;

  /* Memory allocation for x, y, correlation */
  x = (double*)malloc(sizeof(double) * n);
  y = (double*)malloc(sizeof(double) * n);
  correlation = (double*)malloc(sizeof(double) * n);

  /* Initialize x, y, correlation */
  for (i=0; i<n; i++) {
    x[i] = y[i] = i + id * n;
    correlation[i] = 0;
  }

  /* Formular to get left neighbor and right neighbor */
  leftnb = (id - 1 + size) % size;
  rightnb = (id + 1) % size;

  /* Main loops to get evaluation of function 
     Better Message Parallel Algorithm */
  for (k=0; k<size; k++) {
    for (j=0; j<n; j++){
      for (i=0; i<n; i++){
        correlation[i] += (double)(x[i]*x[i] + y[j]*y[j]);
      }
    MPI_Send(y, n, MPI_DOUBLE, leftnb, 1, ALL);
    MPI_Recv(y, n, MPI_DOUBLE, rightnb, 1, ALL, &status);
    }
  }
 
  /* Sumarize all correlation */
  for (i=1; i<n; i++){
    correlation[0] += correlation[i];
  }

  /* The result is allocated in the process 0 */
  if (id == 0) {
    for (i=1; i<size; i++) {
      MPI_Recv(&buffer, 1, MPI_DOUBLE, i, 1, ALL, &status);
      correlation[0] += buffer;
    }
    printf("The result is: %g\n", correlation[0]);
  }
  else {
    MPI_Send(&correlation[0], 1, MPI_DOUBLE, 0, 1, ALL);
  }

  /* Finalize MPI functions */
  MPI_Finalize();
}
